;;
;;  Lambda Helper
;;
;;  When using lambda through an Emacs *shell* buffer (or using M-x
;;  lambda), this adds macro completion and the M-i feature of
;;  `lambda' itself.
;;
;;  M-i leads to a hypertext help system:
;;    RET follows cross references,
;;    i shows an index of macro names,
;;    l goes back to previous page viewed
;;

(require 'cl)

(make-local-variable 'lambda-active)

(defvar lambda-program "./lambda")

(defvar lambda-special-help
  '(("Index" lambda-get-help-index))
  "Association of words / help

Each car ist a string, cadr is a function that returns the
help text")

(defvar lambda-help-history nil)

(defface lambda-help-heading-face '((t (:bold t)))
  "Face used for heading in help windows")

(defface lambda-help-xref-face '((t (:underline t)))
  "Face used for cross references in help windows")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Interface to lambda process
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun lambda-output-filter (PROCESS STRING)
  ;; This output filter allows to capture `lambda's responses to
  ;; a command.
  (if lambda-command-running
      (progn
	(let ((M (string-match "lambda>" STRING)))
	  (if M
	      (setq lambda-command-running nil
		    STRING (substring STRING 0 M)))
	  (setq lambda-command-output (concat lambda-command-output STRING))
	  )
	)
    (if lambda-old-process-filter
	(funcall lambda-old-process-filter PROCESS STRING))
    )
  )

(defun lambda-input-filter (STRING)
  ;; Recognizes certain command names...
  (if (string-match "^\\(def\\|.def\\|read\\|undef\\)\\>" STRING)
      (setq lambda-need-update-macro-list t)
    (if (string-match "^\\exit\\>" STRING)
	(disable-lambda)
      )
    )
  )

(defun lambda-process-command (CMD)
  "Process the command using lambda

Submits the specified command to lambda. The command must
include the terminating newline. Returns lambda's response
as a string. Must be called from *lambda* buffer."
  (if (not lambda-active)
      (error "lambda not active"))
  (setq lambda-command-running t
	lambda-command-output "")
  (process-send-string (get-buffer-process (current-buffer)) CMD)
  (while lambda-command-running
    (accept-process-output))
  lambda-command-output
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Completion
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun lambda-update-macro-list ()
  "Update cache of lambda's macro names

For completing macro names, we keep a list of currently
defined macros in `lambda-macro-list-cache'. Use this command
to update the list (this sends a `show ALL' command to lambda
and parses the output). Usually, this is done automatically
when needed."
  (interactive)
  (let ((WHAT (lambda-process-command "show ALL;\n"))
	VALUE LAST)
    (while (string-match "\n\\(<[^>]*>\\)" WHAT LAST)
      (setq VALUE (cons (match-string 1 WHAT) VALUE))
      (setq LAST (match-end 0)))
    (setq lambda-macro-list-cache (sort VALUE 'string<))
    (setq lambda-need-update-macro-list nil)
    )
  )

(defun lambda-update-command-list ()
  "Update cache of lambda's command names"
  (interactive)
  (let ((WHAT (lambda-process-command "help;\n"))
	VALUE LAST)
    (while (string-match "\n\\(\\w+\\)[^\n]* - " WHAT LAST)
      (setq VALUE (cons (match-string 1 WHAT) VALUE))
      (setq LAST (match-end 0)))
    (setq lambda-command-list-cache VALUE)
    )
  )

(defun lambda-update-option-list ()
  "Update cache of lambda's option names"
  (interactive)
  (let ((WHAT (lambda-process-command "set;\n"))
	VALUE LAST)
    (while (string-match "\n\\([^ ]+\\) = " WHAT LAST)
      (setq VALUE (cons (match-string 1 WHAT) VALUE))
      (setq LAST (match-end 0)))
    (setq lambda-option-list-cache VALUE)
    )
  )

(defun lambda-get-macro-list ()
  "Get list of macros currently defined in lambda

Returns a list of macros currently defined in lambda,
such as `(list \"<foo>\" \"<bar>\")'. Uses a cache which
is only updated when needed (`lambda-need-update-macro-list')."
  (if lambda-need-update-macro-list
      (lambda-update-macro-list))
  lambda-macro-list-cache
  )

(defun lambda-get-command-list ()
  "Return list of valid lambda commands"
  (if (not lambda-command-list-cache)
      (lambda-update-command-list))
  lambda-command-list-cache
  )

(defun lambda-get-option-list ()
  "Return list of valid lambda options"
  (if (not lambda-option-list-cache)
      (lambda-update-option-list))
  lambda-option-list-cache
  )

(defun lambda-complete-macro ()
  ;; if the user attempts to complete a macro, set FUNC/PREFIX
  ;; appropriately and return non-nil.
  (save-excursion
    (let (L LEN)
      (setq L (skip-chars-backward "^ \n<>?:.;\\\\"))
      (if (and L (not (zerop (skip-chars-backward "<"))))
	  ;; It's a macro.
	  (setq LEN (+ 1 (- L))
		FUNC 'lambda-get-macro-list
		PREFIX (buffer-substring (point) (+ (point) LEN)))
	)
      )
    )
  )

(defun lambda-complete-command ()
  ;; if the user attempts to complete a command, set FUNC/PREFIX
  ;; appropriately and return non-nil
  (save-excursion
    (let ((OPOINT (point))
	  L TEXT)
      (setq L (skip-chars-backward "^\n;"))
      (if (looking-at "\\(lambda>\\)? +")
	  (goto-char (match-end 0)))
      (setq TEXT (buffer-substring (point) OPOINT))
      (cond ((string-match "^[^ \t]*$" TEXT)
	     ;; It's a command
	     (setq FUNC 'lambda-get-command-list
		   PREFIX TEXT))
	    ((string-match "^set +\\([^ \t]*\\)$" TEXT)
	     ;; It's a `set' option
	     (setq FUNC 'lambda-get-option-list
		   PREFIX (match-string 1 TEXT)))
	    )
      )
    )
  )

(defun lambda-completion-function ()
  "Lambda completion function

Completes lambda macro names. This function is
plugged into `comint-dynamic-complete-functions'."
  (let (FUNC PREFIX)
    (or (lambda-complete-macro)
	(lambda-complete-command))
    (if FUNC
	(comint-dynamic-simple-complete PREFIX (funcall FUNC))
      )
    )
  )

(defun lambda-get-help-index ()
  "Obtain help index

Can be called from a *lambda* or *lambda-help* buffer.
Returns the text for the help index."
  (save-excursion
    (if (boundp 'lambda-this-is-a-help-buffer)
	(set-buffer lambda-this-is-a-help-buffer))
    (mapconcat 'identity (lambda-get-macro-list) "\n"))
  )

(defun lambda-get-macro-help (NAME)
  "Obtain help for macro NAME

Can be called from a *lambda* or *lambda-help* buffer.
Returns the result of \"show NAME;\" in lambda, or nil
if the macro is not defined (see `lambda-get-macro-list')."
  (save-excursion
    (let ((A (assoc NAME lambda-special-help)))
      (if A (funcall (cadr A))
	(if (boundp 'lambda-this-is-a-help-buffer)
	    (set-buffer lambda-this-is-a-help-buffer))
	(and (member NAME (lambda-get-macro-list))
	     (setq TEXT (lambda-process-command (concat "show " NAME ";\n"))))
	)
      )
    )
  )

(defun lambda-show-help-text (NAME TEXT)
  "Show lambda help text

Can be called from *lambda* or *lambda-help* buffer.
Show a help text with heading NAME and text TEXT."
  ;; If this is not the help buffer, create/switch to one, and
  ;; set up the local variables / keymap.
  (if (not (boundp 'lambda-this-is-a-help-buffer))
      (let ((BUFFER (current-buffer)) 
	    (NEWBUF (get-buffer-create (concat "*lambda-help*"))))
	(split-window)
	(other-window 1)
	(switch-to-buffer NEWBUF)
	(help-mode)
	(make-variable-buffer-local 'lambda-this-is-a-help-buffer)
	(setq lambda-this-is-a-help-buffer BUFFER)
	(use-local-map lambda-help-mode-map)
	)
    )

  ;; make help text
  (toggle-read-only -1)
  (delete-region (point-min) (point-max))
  (insert NAME "\n\n")
  (set-extent-face (make-extent (point-min) (point-max)) 
		   'lambda-help-heading-face)
  (save-excursion
    (insert TEXT))
  (save-excursion
    (while (search-forward-regexp "<[^>=\n ]+>" nil t)
      (set-extent-face (make-extent (match-beginning 0) (match-end 0))
		       'lambda-help-xref-face)))

  (setq lambda-help-history (cons NAME lambda-help-history))
  (toggle-read-only)
  )

(defun lambda-describe-macro (ASK)
  "Show information for a lambda macro

Can be called from a *lambda* or *lambda-help* buffer.
Shows information for the macro the cursor is on, or
asks for the macro name if called with prefix argument
(i.e. ARG non-nil)."
  (interactive "P")
  (let (NAME TEXT)
    ;; Obtain macro name
    (if ASK
	;; ask if called with prefix arg
	(progn
	  (setq NAME (read-string "Macro name: "))
	  (if NAME
	      ;; user can enter macro name either with angle brackets or not
	      (or (string-match "^<[^>]*>$" NAME)
		  (setq NAME (concat "<" NAME ">")))))
      (save-excursion
	;; attempt to extract macro name from current buffer
        (if (not (looking-at "<"))
            (skip-chars-backward " \t\n>"))
	(if (and (skip-chars-backward ">")
		 (skip-chars-backward "^ \t\n<>")
		 (skip-chars-backward "<")
		 (looking-at "<[^>]+>"))
	    (setq NAME (match-string 0)))
	)
      )
    (if NAME
	;; OK, there is a name. Get the help text        emacs fontifies this wrong, so -> "
	(if (setq TEXT (lambda-get-macro-help NAME))
            (lambda-show-help-text NAME TEXT)
	  (message "Macro %s not defined." NAME)
	  )
      )
    )
  )

(defun lambda-show-help-index ()
  "Show the lambda help index

Can be called from *lambda* or *lambda-help* buffer."
  (interactive)
  (lambda-show-help-text "Index" (lambda-get-macro-help "Index"))
  )

(defun lambda-last-help ()
  "Show previous lambda help page"
  (interactive)
  (if (and lambda-help-history (cdr lambda-help-history))
      (let (NAME TEXT)
	(setq NAME (cadr lambda-help-history))
	(setq TEXT (lambda-get-macro-help NAME))
	(setq lambda-help-history (cddr lambda-help-history))
	(if TEXT
	    (lambda-show-help-text NAME TEXT)
	  (lambda-last-help))
	)
    (message "No previous help page.")
    )
  )

(defun enable-lambda ()
  "Activate lambda support

When invoking lambda in a *shell* buffer, use this command.
To run lambda in a dedicated buffer (preferred), use
\\[lambda].

This will allow you to complete macro names `<foo>' using
the tab key, as usual."
  (interactive)
  (if (and (boundp 'lambda-active)
	   lambda-active)
      (error "lambda already active"))
  (make-local-variable 'lambda-command-running)
  (make-local-variable 'lambda-command-output)
  (make-local-variable 'lambda-old-process-filter)
  (make-local-variable 'lambda-need-update-macro-list)
  (make-local-variable 'lambda-macro-list-cache)
  (make-local-variable 'lambda-option-list-cache)
  (make-local-variable 'lambda-command-list-cache)
  (make-local-variable 'lambda-active)
  (setq lambda-command-running nil)
  (setq lambda-command-output "")
  (setq lambda-old-process-filter (process-filter (get-buffer-process (current-buffer))))
  (if lambda-old-process-filter
      (set-process-filter (get-buffer-process (current-buffer)) 'lambda-output-filter)
    (error "no process filter active -- can't handle that"))
  (setq comint-dynamic-complete-functions (cons 'lambda-completion-function comint-dynamic-complete-functions))
  (setq comint-input-filter-functions (cons 'lambda-input-filter comint-input-filter-functions))
  (setq comint-prompt-regexp "^\\( +\\|lambda\\)> +")
  (setq lambda-need-update-macro-list t)
  (setq lambda-active t)
  (if (and (boundp 'lambda-mode-map)
	   lambda-mode-map)
      nil
    (setq lambda-mode-map (make-sparse-keymap))
    (set-keymap-name lambda-mode-map 'lambda-mode-map)
    (set-keymap-parents lambda-mode-map (list comint-mode-map))
    (define-key lambda-mode-map "\t" 'comint-dynamic-complete)
    (define-key lambda-mode-map "\M-?" 'comint-dynamic-list-completions)
    (define-key lambda-mode-map "\M-i" 'lambda-describe-macro)
    )
  (if (and (boundp 'lambda-help-mode-map)
	   lambda-help-mode-map)
      nil
    (setq lambda-help-mode-map (make-sparse-keymap))
    (set-keymap-name lambda-help-mode-map 'lambda-help-mode-map)
    (set-keymap-parents lambda-help-mode-map (list help-mode-map))
    (define-key lambda-help-mode-map "\n" 'lambda-describe-macro)
    (define-key lambda-help-mode-map "\r" 'lambda-describe-macro)
    (define-key lambda-help-mode-map "\M-i" 'lambda-describe-macro)
    (define-key lambda-help-mode-map "i" 'lambda-show-help-index)
    (define-key lambda-help-mode-map "l" 'lambda-last-help)
    )
  (use-local-map lambda-mode-map)

  (setq font-lock-keywords '(("^lambda>" 0 bold)
			     ("^[0-9]+\\[[0-9,SKIen]+\\]:" 0 yellow)))
  )

(defun disable-lambda ()
  "Disable lambda support in *shell* buffer again

This function is automatically called when a `exit' command is
recongnized."
  (interactive)
  (if lambda-active
      (progn
	(set-process-filter (get-buffer-process (current-buffer)) lambda-old-process-filter)
	(setq comint-input-filter-functions (remove 'lambda-input-filter comint-input-filter-functions))
	(setq comint-dynamic-complete-functions (remove 'lambda-completion-function comint-dynamic-complete-functions))
	(setq lambda-active nil))
    )
  )

(defun lambda ()
  "Run lambda in a buffer

This command runs lambda in a buffer, like \\[shell] does with
the shell. The location of the lambda program must be set in
the variable `lambda-program'.

See `enable-lambda' for what benefits this command gives you."
  (interactive)
  (switch-to-buffer "*lambda*")
  (make-comint "lambda" lambda-program)
  (enable-lambda)
  )
