// glaub mir, emacs, das ist -*- c++ -*-
//
//  Lambda
//  . Global benutzte Dinge
//  . CNode (Basisklasse)
//  . CLambda (Reprsentation von lambda x.term)
//
//  (c) copyright 1998,2000 by Stefan Reuther <Streu@gmx.de>
//

#ifndef NODE_H
#define NODE_H

#include <iostream>
#include <string>
#include <map>

#include "type.h"

// Debugmakro
//#define WHERE std::cerr<<__PRETTY_FUNCTION__<<std::endl
#define WHERE

// Liste der freien Variablen
struct TVariable {
    TVariable* next;
    bool marker;		// Markierung fr freevars()
    std::string name;
    std::string replaceby;
    TVariable(TVariable* anext, std::string& aname) : next(anext),
	marker(false), name(aname), replaceby() {}
};

// Typinferenz -- Typ einer gebundenen Variablen
struct TTypeItem {
    TTypeItem* next;
    std::string name;
    CTypePtr type;
};

// Typinferenz -- Statusblock
struct TTypeInfo {
    std::string error_msg;
    bool annotate;
    std::map<std::string, CTypePtr> free_types; // Typen der freien Variablen
};
  
// Parameterblock fr numcheck()
struct TNumCheck {
    int value;
    enum { CHECK_X, CHECK_Y, CHECK_RIGHT, CHECK_LEFT } status;
    std::string x, y;
};

struct TListCheck {
    typedef enum { CHECK_F, CHECK_APP, CHECK_1ST_APP, CHECK_1ST,
		   CHECK_2ND, CHECK_TAIL, CHECK_NIL } TStatus;
    TStatus status;
    int length;
    std::string f;
    std::string nil;
};

typedef enum { PRINT_UNDEF,          // immer klammern
	       PRINT_LEFT_APP,       // Linke Seite einer Applikation
	       PRINT_RIGHT_APP,      // Rechte Seite
	       PRINT_ABSTRACT,       // ?x.das_hier
	       PRINT_LAMBDA          // das y in ?x y.
} TPrintContext;

struct TIoContext {
    TPrintContext pc;
    int parens_needed;          // noch zu schlieende Klammern
};

class CLambda;
class CSymbol;
class CApplication;

//
//  Ein Objekt im "Termbaum"
//
class CNode {
public:
    CNode() {}
    virtual ~CNode() {}

    // Diesen Knoten reduzieren. Liefert neuen Knoten (Ersatz fr
    // diesen), dieser Knoten wird u.U. invalidiert (`(M N)' wird
    // zu etwas in der Art von `(M #null)' bei Rckgabe von
    // `(M' N)' (Reduktion in M angewandt, normal order)
    virtual CNode* reduce() = 0;
  	
    // Liefert einen Zeiger auf ein CLambda, falls der Knoten einem
    // "lambda X.expr" entspricht, sonst NULL.
    virtual CLambda* lambda() { return 0; }

    // Test fr Eta-Konversion:
    // stage=0: Test auf (M x); Rckgabe ist Kopie von M oder 0
    // stage=1: Test auf x;     Rckgabe ist x oder 0
    // im Erfolgsfall ist der Baum invalidiert (i.e., `(#null x)')
    virtual CNode* eta(int stage, std::string name) { return 0; }
  	
    // Liefert eine Kopie des Baumes
    virtual CNode* copy() = 0;
  	
    // Druckt den Baum aus
    virtual void print(std::ostream& out, TPrintContext pc=PRINT_UNDEF) = 0;
  	
    // Modifiziert Liste der freien Variablen entsprechend
    virtual TVariable* freevars(TVariable* p) = 0;
  	
    // Alpha-Konversion, so da die Variablen in fv nicht mehr
    // gebunden sind.
    virtual bool alpha(TVariable* fv) = 0;

    // Globale Alpha-Konversion
    // action < 0: alle Variablen anonymisieren
    // action >= 0: alle anonymen Variablen zurckersetzen, wobei
    //   action ein Index in den "Nutzer-Namensraum" ist:
    //   0->a, 1->b, .. 25->z
    virtual void global_alpha(TVariable* fv, int action) = 0;
  	
    // Liefert Kopie des Baumes, in dem Referenzen auf sym durch
    // rep->copy() ersetzt werden.
    virtual CNode* replace(std::string& sym, CNode* rep) = 0;

    // Term in SKI-Form bringen
    virtual CNode* ski() = 0;

    // Term in SKI-Form bringen, der das `X' in `?name.X' ist
    // X mu bereits in SKI-Form sein
    virtual CNode* ski_lambda(std::string name) = 0;
    
    // Prft, ob es sich um die numerische Reprsentation einer
    // Zahl handelt
    virtual bool numcheck(TNumCheck& nc) = 0;
    virtual bool listcheck(TListCheck& lc) = 0;

    virtual void listout(std::ostream& out, int n, char delim) { print(out); }

    // Typ inferieren
    virtual CTypePtr get_type(TTypeInfo& ti, TTypeItem* local_types) = 0;

    // Anzahl Atome zhlen
    virtual int count_atoms() { return 1; }
    
    // Wrapper fr reduce, wirft das Original weg
    static CNode* do_reduce(CNode*&p)
	{ CNode* r = p->reduce(); delete p; p = 0; return r; }
  	 
    // Wrapper fr replace, wirft das Original weg
    static CNode* do_replace(CNode*&p, std::string& sym, CNode* rep)
	{ CNode* r = p->replace(sym,rep); delete p; p = 0; return r; }

    // Reduzieren mit gleichzeitiger Ausgabe. Liefert einen neuen
    // Knoten, invalidiert diesen u.U.
    virtual CNode* ioreduce(std::ostream& os, TIoContext& ioc);

    // SKI-Reduktion. Liefert neuen Knoten, invalidiert diesen u.U.
    virtual CNode* ski_reduce();

    // Liefert Zeiger auf CApplication-Objekt, falls dies eine
    // Applikation ist, sonst 0
    virtual CApplication* applic() { return 0; }

    // true gdw. dies der Kombinator c ist
    virtual bool is_ski(char c) { return false; }

    // true gdw. dieser Ausdruck das angegebene Makro (direkt
    // oder indirekt) aufruft
    virtual bool calls_macro(std::string);

    // Makrorekursions-Elimination: Ersetzt Aufrufe von Makro mac
    // durch Bezeichner id
    virtual CNode* replace_macro(std::string mac, std::string id);
};

//
//  Ausdruck der Form "lambda symbol:typ.sub"
//
class CLambda : public CNode {
    std::string symbol;
    CNode *sub;
    CTypePtr type;
public:
    CLambda(const std::string& asym, CNode* asub, CTypePtr atype)
        : symbol(asym), sub(asub), type(atype) {}
    virtual ~CLambda() { delete sub; }
  	
    virtual CNode* reduce();
    virtual CLambda* lambda() { return this; }
    virtual CNode* copy();
    virtual void print(std::ostream& out, TPrintContext pc);
    virtual TVariable* freevars(TVariable* p);
    virtual bool alpha(TVariable* fv);
    virtual void global_alpha(TVariable* fv, int action);
    virtual CNode* replace(std::string& sym, CNode* rep);
    virtual bool numcheck(TNumCheck& nc);
    virtual bool listcheck(TListCheck& lc);
    virtual void listout(std::ostream& out, int n, char delim);

    virtual int count_atoms() { return sub->count_atoms() + 1; }
    virtual CTypePtr get_type(TTypeInfo& ti, TTypeItem* local_types);

    virtual CNode* ski();
    virtual CNode* ski_lambda(std::string);

    // liefert sub[symbol/rep] (Einsetzen)
    virtual CNode* transform(CNode* rep);
    
    virtual CNode* ioreduce(std::ostream& os, TIoContext& ioc);

    virtual bool calls_macro(std::string);
    virtual CNode* replace_macro(std::string mac, std::string id);
};

enum TReductionOrder {
    LAZY_REDUCTION,             // leftmost-outermost
    BROAD_REDUCTION,            // "Breitenreduktion"
    EAGER_REDUCTION,            // leftmost-innermost

    REDUCTIONORDER_MAX
};
  
// Hilfsmittel
std::string newtemp();
extern int tempnum;
extern bool changed;
extern bool full_paren_flag;
extern bool newlist_flag;       // neues Listenformat (?nil. ...)
extern bool true_ok;            // wahr, wenn Makro <true> definiert und OK
extern bool pre_alpha_flag;
extern bool post_alpha_flag;
extern bool ski_shortcut_flag;  // ?x.<unabhngig-von-x>  =>*  K <u-v-x>

extern TReductionOrder reduction_order;

extern int n_beta, n_eta;       // Statistik ber Anzahl Konversionen
extern int n_expand;

void deletelist(TVariable* pv);
bool try_print_list(std::ostream& o, CNode* p, bool report=false);

// Termbaum umhngen (invalidiert Quelle)
inline CNode* move(CNode*& src)
{
    // Original: return src->copy();
    CNode* val = src;
    src = 0;
    return val;
}

#endif
