//
//  Lambda
//  . Typen
//
//  (c) copyright 2000 by Stefan Reuther <Streu@gmx.de>
//

/* CType-Speichermanager debuggen? */
#undef DEBUG_TYPES

#include <iostream>
#include <cstdio>
#include "type.h"
#include "main.h"
#include "node.h"

#ifdef DEBUG_TYPES
#include <set>
#endif

using namespace std;

int CType::n_types = 0;

#ifdef DEBUG_TYPES
static set<CType*> registry;
void type_reg_print()
{
    for(set<CType*>::iterator i = registry.begin();
        i != registry.end();
        i++) {
        cout << "## " << (*i)->to_string() << endl;
    }
}

#  define REGISTER_TYPE registry.insert(this);
#else
#  define REGISTER_TYPE
#endif

/* Funktionstyp `ain -> aout' */
CType::CType(CTypePtr ain, CTypePtr aout)
    : type(TYPE_FUNCTION), ref(0), in(ain), out(aout), name(),
      rec_flags(0)
{
    n_types++;
    REGISTER_TYPE;
}

/* Typalias `ain' */
CType::CType(CTypePtr ain)
    : type(TYPE_ALIAS), ref(0), in(ain), out(0), name(),
      rec_flags(0)
{
    n_types++;
    REGISTER_TYPE;
}

/* Typatom `aname' */
CType::CType(std::string aname)
    : type(TYPE_ATOM), ref(0), in(0), out(0), name(aname),
      rec_flags(0)
{
    n_types++;
    REGISTER_TYPE;
}

/* unbekannter Typ (Platzhalter) */
CType::CType()
    : type(TYPE_UNKNOWN), ref(0), in(0), out(0), name(),
      rec_flags(0)
{
    n_types++;
    REGISTER_TYPE;
}

CType::~CType()
{
    n_types--;
#ifdef DEBUG_TYPES
    registry.erase(this);
#endif
}

/* Typ in String konvertieren. Bei !with_paren werden die Klammern
   um Funktionstypen weggelassen, falls mglich */
string CType::to_string(bool with_paren)
{
    if(rec_flags & REC_TOSTRING)
        return "[Rekursion]";
    
    CProtector prot(rec_flags, REC_TOSTRING);

    if(full_paren_flag)
        with_paren = true;
    
    switch(type) {
     case TYPE_ATOM:
        return name;

     case TYPE_FUNCTION:
        {
            string ret;
            if(with_paren)
                ret = "(";
            ret += in->to_string(true);
            ret += " -> ";
            ret += out->to_string(false);
            if(with_paren)
                ret += ")";
            return ret;
        }

     case TYPE_ALIAS:
        return in->to_string(with_paren);
        
     case TYPE_UNKNOWN:
        {
            char buf[50];
            sprintf(buf, "#%p", (void*)this);
            return buf;
        }
        
     default:
        return "{Huh?}";
    }
}

/* Funktionstyp erzeugen */
void CType::make_function(CTypePtr ain, CTypePtr aout)
{
    type = TYPE_FUNCTION;
    in = ain;
    out = aout;
}

/* Alias auf other erzeugen */
void CType::make_alias(CTypePtr other)
{
    if(other->get_ptr() == get_ptr())
        return;
    
    type = TYPE_ALIAS;
    in = other;
    out = 0;
}

/*
 *  Zwei Typen "unifizieren" (angleichen). Liefert true wenn das mglich
 *  ist, sonst false und in /error_msg/ eine Fehlermeldung.
 *
 *  Es mu hllisch aufgepat werden, da keine rekursiven Typen
 *  entstehen. Da /CType/s referenzgezhlt verwaltet werden, ergibt
 *  ein rekursiver Typ ein Speicherleck.
 */
bool CType::unify(CTypePtr other, string& error_msg)

{
    if(rec_flags & REC_UNIFY) {
        error_msg = "Selbstapplikation / Rekursiver Typ";
        return false;
    }
    if(ctrlcpressed) {
        error_msg = "Abbruch in unify";
        return false;
    }
    
    other.sync();

    CProtector prot(rec_flags, REC_UNIFY);
    if(type != TYPE_ALIAS && other->type == TYPE_UNKNOWN) {
        /* alles pat auf UNKNOWN: UNKNOWN umaliasen */
        if(other->get_ptr() == get_ptr())
            return true;
        if(contains(other->get_ptr()) || other->contains(this)) {
            error_msg = "Rekursiver Typ [1]";
            return false;
        } else {
            other->make_alias(this);
            return true;
        }
    }
    
    switch(type) {
     case TYPE_ALIAS:
        /* Aliase bergehen */
        return in->unify(other, error_msg);
        
     case TYPE_ATOM:
        /* Atome mssen passen */
        if(other->type == TYPE_ATOM && other->name == name) {
            return true;
        } else {
            error_msg = "Typkonflikt: `" + name + "' vs. `" + other->to_string() + "'";
            return false;
        }
                
     case TYPE_FUNCTION:
        /* Funktionstypen mssen passen */
        if(other->type != TYPE_FUNCTION) {
            error_msg = "Typkonflikt: `" + to_string() + "' vs. `" + other->to_string() + "'";
            return false;
        }

        return (in->unify(other->in, error_msg) &&
                out->unify(other->out, error_msg));
        
     case TYPE_UNKNOWN:
        /* unbekannt versus irgendwas: */
        if(other->get_type() == TYPE_UNKNOWN)
            cerr << "can't happen -- 2x TYPE_UNKNOWN" << endl;
        
        if(other->contains(this)) {
            error_msg = "Rekursiver Typ [2]";
            return false;
        } else {
            make_alias(other);
            return true;
        }
    }
    error_msg = "Ungltiger Typ-Knoten";
    return false;
}

static char* names[] = {
    "Alpha", "Beta",  "Gamma", "Delta",
    "Eps",   "Zeta",  "Eta",   "Theta",
    "Iota",  "Kappa", "My",    "Ny",
    "Xi",    "Omikr", "Pi",    "Rho",
    "Sigma", "Tau",   "Yps",   "Phi",
    "Chi",   "Psi",   "Omega"
};

/* Liefert einen benutzerfreundlichen Namen korrespondierend mit Slot N */
static string get_user_type_name(int n)
{
    const int modulus = sizeof(names) / sizeof(names[0]);
    
    char buffer[50];
    int num = n / modulus;
    if(num) {
        sprintf(buffer, "%s%d", names[n % modulus], num);
        return buffer;
    } else
        return names[n % modulus];
}

/*
 *  Unbekannte Typknoten umbenennen fr die Ausgabe.
 */
void CType::rename_unknown(int& n)
{
    if(rec_flags & REC_RENAME)
        return;
    CProtector(rec_flags, REC_RENAME);
    
    switch(type) {
     case TYPE_ALIAS:
        in->rename_unknown(n);
        break;
     case TYPE_FUNCTION:
        in->rename_unknown(n);
        out->rename_unknown(n);
        break;
     case TYPE_UNKNOWN:
        name = get_user_type_name(n++);
        type = TYPE_ATOM;
        break;
     case TYPE_ATOM:
        break;
    }
}

/* true iff dieser Typ ist in other enthalten */
bool CType::contains(CType* other)
{
    while(other->type == TYPE_ALIAS)
        other = other->in->get_ptr();
    
    if(other == this)
        return true;

    switch(type) {
     case TYPE_ALIAS:
        return in->contains(other);
     case TYPE_FUNCTION:
        return in->contains(other) || out->contains(other);
     case TYPE_UNKNOWN:
     case TYPE_ATOM:
        break;
    }
    return false;
}
