//
//  Lambda
//  . Typen
//
//  (c) copyright 2000 by Stefan Reuther <Streu@gmx.de>
//
#ifndef TYPE_H
#define TYPE_H

#include <string>
#include <iosfwd>

class CType;

/* kapselt einen Zeiger auf einen Typ */
class CTypePtr {
    CType* ptr;
    /* Referenzzhler managen */
    void inc() const;
    void dec() const;
 public:
    CTypePtr(CType* aptr = 0) : ptr(aptr) { inc(); }
    ~CTypePtr() { dec(); ptr = 0; }

    CTypePtr(const CTypePtr& other) : ptr(other.ptr) {
        inc();
    }
    CTypePtr& operator=(CType* p);
    CTypePtr& operator=(const CTypePtr& other) {
        if(other.ptr != ptr) {
            other.inc();
            dec();
            ptr = other.ptr;
        }
        return *this;
    }

    CType& operator*() { return *ptr; }
    CType* operator->() { return ptr; }
    void sync();
    
    operator bool() { return ptr; }
};

/* Hilfsklasse fr Rekursionsschutz */
struct CProtector {
    int& what;
    int mask;
    CProtector(int& awhat, int amask)
        : what(awhat), mask(amask) { what |= mask; }
    ~CProtector() { what &= ~mask; }
};

/* Typ */
class CType {
 public:
    enum Type {
        TYPE_ATOM,              // Typ-Atom /name/
        TYPE_FUNCTION,          // Funktion /in -> out/
        TYPE_UNKNOWN,           // unbekannter Typ
        TYPE_ALIAS              // Alias auf /in/
    };
    enum {
        REC_TOSTRING = 1,
        REC_UNIFY    = 2,
        REC_RENAME   = 4
    };
    static int n_types;         // Anzahl Typen im System (zur Kontrolle)
 private:
    Type type;
    int ref;                    // Referenzzhler
    CTypePtr in, out;           // Untertypen
    std::string name;           // Atom-Name
    int rec_flags;              // Rekursionserkennung
    
    /* Keiner, der mich je gesehen, htte das geglaubt: Kopieren ist bei
       mir nicht erlaubt. */
    CType(const CType&);
    CType& operator=(const CType&);

    friend class CTypePtr;
 public:
    CType(CTypePtr ain, CTypePtr aout);  // Funktionstyp
    CType(CTypePtr ain);                 // Typalias
    CType(std::string aname);            // Atom
    CType();                             // unbekannt
    ~CType();

    std::string to_string(bool with_paren = true);

    /* Typaliase auflsen */
    CType* get_ptr() {
        if(type==TYPE_ALIAS)
            return in->get_ptr();
        else
            return this;
    }

    Type get_type() { return type; }
    std::string get_name() { return name; }
    CTypePtr get_in() { return in; }
    CTypePtr get_out() { return out; }
    
    void make_function(CTypePtr ain, CTypePtr aout);
    void make_alias(CTypePtr other);
    bool contains(CType* other);
    bool unify(CTypePtr other, std::string& errmsg);

    void rename_unknown(int& n);
};

/* intern: Referenzzhler erhhen */
inline void CTypePtr::inc() const
{
    if(ptr) ptr->ref++;
}

/* intern: Referenzzhler erniedrigen */
inline void CTypePtr::dec() const
{
    if(ptr && !--ptr->ref) delete ptr;
}

/* Aliase auflsen */
inline void CTypePtr::sync()
{
    if(ptr)
        operator=(ptr->get_ptr()); // operator= pat auf Referenzzhler auf
}

inline CTypePtr& CTypePtr::operator=(CType* p)
{
    if(p)
        p->ref++;
    dec();
    ptr = p;
    return *this;
}

#endif
